from artifact_torch._base.components.callbacks.backward_hook import (
    BackwardHookArrayCallback,
    BackwardHookArrayCollectionCallback,
    BackwardHookCallback,
    BackwardHookPlotCallback,
    BackwardHookPlotCollectionCallback,
    BackwardHookScoreCallback,
    BackwardHookScoreCollectionCallback,
)

__all__ = [
    "BackwardHookArrayCallback",
    "BackwardHookArrayCollectionCallback",
    "BackwardHookCallback",
    "BackwardHookPlotCallback",
    "BackwardHookPlotCollectionCallback",
    "BackwardHookScoreCallback",
    "BackwardHookScoreCollectionCallback",
]



from typing import Any, Generic, TypeVar

from artifact_core.typing import (
    Array,
    ArrayCollection,
    ArtifactResult,
    Plot,
    PlotCollection,
    Score,
    ScoreCollection,
)

from artifact_torch._base.components.handlers.hook import HookCallbackHandler
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
CacheDataTCov = TypeVar("CacheDataTCov", bound=ArtifactResult, covariant=True)


class ModelIOCallbackHandler(
    HookCallbackHandler[Model[Any, ModelOutputTContr], CacheDataTCov],
    Generic[ModelOutputTContr, CacheDataTCov],
):
    pass


ModelIOScoreHandler = ModelIOCallbackHandler[ModelOutputTContr, Score]


ModelIOArrayHandler = ModelIOCallbackHandler[ModelOutputTContr, Array]


ModelIOPlotHandler = ModelIOCallbackHandler[ModelOutputTContr, Plot]


ModelIOScoreCollectionHandler = ModelIOCallbackHandler[ModelOutputTContr, ScoreCollection]


ModelIOArrayCollectionHandler = ModelIOCallbackHandler[ModelOutputTContr, ArrayCollection]


ModelIOPlotCollectionHandler = ModelIOCallbackHandler[ModelOutputTContr, PlotCollection]



from abc import abstractmethod
from typing import Generic, TypedDict, TypeVar

from artifact_core.shared.collections import ClassificationResults

from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelInputTContr = TypeVar("ModelInputTContr", bound="ModelInput", contravariant=True)
ModelOutputTCov = TypeVar("ModelOutputTCov", bound="ModelOutput", covariant=True)
ClassificationParamsTContr = TypeVar(
    "ClassificationParamsTContr", bound="ClassificationParams", contravariant=True
)
ClassificationDataTContr = TypeVar("ClassificationDataTContr", contravariant=True)
ClassificationResultsTCov = TypeVar(
    "ClassificationResultsTCov", bound=ClassificationResults, covariant=True
)


class ClassificationParams(TypedDict):
    pass


class Classifier(
    Model[ModelInputTContr, ModelOutputTCov],
    Generic[
        ModelInputTContr,
        ModelOutputTCov,
        ClassificationParamsTContr,
        ClassificationDataTContr,
        ClassificationResultsTCov,
    ],
):
    @abstractmethod
    def forward(self, model_input: ModelInputTContr, *args, **kwargs) -> ModelOutputTCov: ...

    @abstractmethod
    def classify(
        self, data: ClassificationDataTContr, params: ClassificationParamsTContr
    ) -> ClassificationResultsTCov: ...



from abc import abstractmethod
from typing import Generic, TypeVar

import pandas as pd

from artifact_torch._base.model.io import (
    ModelInput,
    ModelOutput,
)
from artifact_torch._domains.generation.model import GenerationParams, GenerativeModel

ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTCov = TypeVar("ModelOutputTCov", bound=ModelOutput, covariant=True)
GenerationParamsTContr = TypeVar(
    "GenerationParamsTContr", bound=GenerationParams, contravariant=True
)


class TableSynthesizer(
    GenerativeModel[ModelInputTContr, ModelOutputTCov, GenerationParamsTContr, pd.DataFrame],
    Generic[ModelInputTContr, ModelOutputTCov, GenerationParamsTContr],
):
    @abstractmethod
    def forward(self, model_input: ModelInputTContr, *args, **kwargs) -> ModelOutputTCov: ...

    @abstractmethod
    def generate(self, params: GenerationParamsTContr) -> pd.DataFrame: ...



from artifact_core.binary_classification import (
    BinaryClassificationArrayCollectionType,
    BinaryClassificationArrayType,
    BinaryClassificationPlotCollectionType,
    BinaryClassificationPlotType,
    BinaryClassificationScoreCollectionType,
    BinaryClassificationScoreType,
    BinaryClassSpec,
)
from artifact_core.binary_classification.collections import (
    BinaryClassificationResults,
    BinaryClassStore,
    BinaryDistributionStore,
)
from artifact_experiment.binary_classification import BinaryClassificationPlan

from artifact_torch._domains.classification.model import ClassificationParams
from artifact_torch.binary_classification._experiment import BinaryClassificationExperiment
from artifact_torch.binary_classification._model import BinaryClassifier
from artifact_torch.binary_classification._routine import (
    BinaryClassificationRoutine,
    BinaryClassificationRoutineData,
)

__all__ = [
    "BinaryClassificationArrayCollectionType",
    "BinaryClassificationArrayType",
    "BinaryClassificationPlotCollectionType",
    "BinaryClassificationPlotType",
    "BinaryClassificationScoreCollectionType",
    "BinaryClassificationScoreType",
    "BinaryClassSpec",
    "BinaryClassificationResults",
    "BinaryClassStore",
    "BinaryDistributionStore",
    "BinaryClassificationPlan",
    "ClassificationParams",
    "BinaryClassificationExperiment",
    "BinaryClassifier",
    "BinaryClassificationRoutine",
    "BinaryClassificationRoutineData",
]



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Mapping, Optional, Type, TypeVar

from artifact_core.binary_classification.collections import BinaryClassStore
from artifact_core.binary_classification.spi import (
    BinaryClassificationArtifactResources,
    BinaryClassSpecProtocol,
)
from artifact_experiment.binary_classification import BinaryClassificationPlan
from artifact_experiment.tracking import DataSplit

from artifact_torch._base.components.routines.artifact import (
    ArtifactRoutine,
    ArtifactRoutineData,
    ArtifactRoutineHyperparams,
)
from artifact_torch._domains.classification.model import ClassificationParams
from artifact_torch.binary_classification._model import BinaryClassifier

ClassificationParamsTCov = TypeVar(
    "ClassificationParamsTCov", bound=ClassificationParams, covariant=True
)


@dataclass
class BinaryClassificationRoutineHyperparams(
    ArtifactRoutineHyperparams, Generic[ClassificationParamsTCov]
):
    classification_params: ClassificationParamsTCov


ClassificationDataTCov = TypeVar("ClassificationDataTCov", covariant=True)


@dataclass
class BinaryClassificationRoutineData(ArtifactRoutineData, Generic[ClassificationDataTCov]):
    true_class_store: BinaryClassStore
    classification_data: ClassificationDataTCov


ClassificationDataTContr = TypeVar("ClassificationDataTContr", contravariant=True)
BinaryClassificationRoutineT = TypeVar(
    "BinaryClassificationRoutineT", bound="BinaryClassificationRoutine"
)


class BinaryClassificationRoutine(
    ArtifactRoutine[
        BinaryClassifier[Any, Any, ClassificationParamsTCov, Any],
        BinaryClassificationRoutineData[ClassificationDataTContr],
        BinaryClassificationRoutineHyperparams[ClassificationParamsTCov],
        BinaryClassificationArtifactResources,
        BinaryClassSpecProtocol,
    ],
    Generic[ClassificationParamsTCov, ClassificationDataTContr],
):
    @classmethod
    @abstractmethod
    def _get_period(
        cls,
        data_split: DataSplit,
    ) -> Optional[int]: ...

    @classmethod
    @abstractmethod
    def _get_artifact_plan(
        cls, data_split: DataSplit
    ) -> Optional[Type[BinaryClassificationPlan]]: ...

    @classmethod
    @abstractmethod
    def _get_classification_params(cls) -> ClassificationParamsTCov: ...

    @classmethod
    def _get_hyperparams(cls) -> BinaryClassificationRoutineHyperparams[ClassificationParamsTCov]:
        classification_params = cls._get_classification_params()
        hyperparams = BinaryClassificationRoutineHyperparams[ClassificationParamsTCov](
            classification_params=classification_params
        )
        return hyperparams

    def _generate_artifact_resources(
        self, model: BinaryClassifier[Any, Any, ClassificationParamsTCov, Any]
    ) -> Mapping[DataSplit, BinaryClassificationArtifactResources]:
        resources_by_split = {}
        for data_split in self._data.keys():
            classification_results = model.classify(
                data=self._data[data_split].classification_data,
                params=self._hyperparams.classification_params,
            )
            resources = BinaryClassificationArtifactResources.from_stores(
                true_class_store=self._data[data_split].true_class_store,
                classification_results=classification_results,
            )
            resources_by_split[data_split] = resources
        return resources_by_split



from typing import Any, TypeVar

from artifact_torch._base.components.callbacks.hook import HookCallbackResources
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelOutput

ModelOutputTCov = TypeVar("ModelOutputTCov", bound=ModelOutput, covariant=True)


ModelIOCallbackResources = HookCallbackResources[Model[Any, ModelOutputTCov]]



from abc import abstractmethod
from typing import Generic, TypedDict, TypeVar

from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelInputTContr = TypeVar("ModelInputTContr", bound="ModelInput", contravariant=True)
ModelOutputTCov = TypeVar("ModelOutputTCov", bound="ModelOutput", covariant=True)
GenerationParamsTContr = TypeVar(
    "GenerationParamsTContr", bound="GenerationParams", contravariant=True
)
GenerationTCov = TypeVar("GenerationTCov", covariant=True)


class GenerationParams(TypedDict):
    pass


class GenerativeModel(
    Model[ModelInputTContr, ModelOutputTCov],
    Generic[ModelInputTContr, ModelOutputTCov, GenerationParamsTContr, GenerationTCov],
):
    @abstractmethod
    def forward(self, model_input: ModelInputTContr, *args, **kwargs) -> ModelOutputTCov: ...

    @abstractmethod
    def generate(self, params: GenerationParamsTContr) -> GenerationTCov: ...



from typing import Any, Dict, Generic, Optional, TypeVar

from artifact_core.typing import (
    Array,
    ArrayCollection,
    Plot,
    PlotCollection,
    Score,
    ScoreCollection,
)

from artifact_torch._base.components.routines.artifact import ArtifactRoutine
from artifact_torch._base.components.routines.loader import DataLoaderRoutine
from artifact_torch._base.components.routines.routine import RoutineResources
from artifact_torch._base.components.routines.train_diagnostics import TrainDiagnosticsRoutine
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)


class RoutineSuite(Generic[ModelTContr, ModelInputTContr, ModelOutputTContr]):
    def __init__(
        self,
        train_diagnostics_routine: Optional[
            TrainDiagnosticsRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]
        ] = None,
        loader_routine: Optional[
            DataLoaderRoutine[ModelTContr, ModelInputTContr, ModelOutputTContr]
        ] = None,
        artifact_routine: Optional[ArtifactRoutine[ModelTContr, Any, Any, Any, Any]] = None,
    ):
        self._train_diagnostics_routine = train_diagnostics_routine
        self._loader_routine = loader_routine
        self._artifact_routine = artifact_routine

    @property
    def scores(self) -> Dict[str, Score]:
        scores = {}
        if self._train_diagnostics_routine is not None:
            scores.update(self._train_diagnostics_routine.scores)
        if self._loader_routine is not None:
            scores.update(self._loader_routine.scores)
        if self._artifact_routine is not None:
            scores.update(self._artifact_routine.scores)
        return scores

    @property
    def arrays(self) -> Dict[str, Array]:
        arrays = {}
        if self._train_diagnostics_routine is not None:
            arrays.update(self._train_diagnostics_routine.arrays)
        if self._loader_routine is not None:
            arrays.update(self._loader_routine.arrays)
        if self._artifact_routine is not None:
            arrays.update(self._artifact_routine.arrays)
        return arrays

    @property
    def plots(self) -> Dict[str, Plot]:
        plots = {}
        if self._train_diagnostics_routine is not None:
            plots.update(self._train_diagnostics_routine.plots)
        if self._loader_routine is not None:
            plots.update(self._loader_routine.plots)
        if self._artifact_routine is not None:
            plots.update(self._artifact_routine.plots)
        return plots

    @property
    def score_collections(self) -> Dict[str, ScoreCollection]:
        score_collections = {}
        if self._train_diagnostics_routine is not None:
            score_collections.update(self._train_diagnostics_routine.score_collections)
        if self._loader_routine is not None:
            score_collections.update(self._loader_routine.score_collections)
        if self._artifact_routine is not None:
            score_collections.update(self._artifact_routine.score_collections)
        return score_collections

    @property
    def array_collections(self) -> Dict[str, ArrayCollection]:
        array_collections = {}
        if self._train_diagnostics_routine is not None:
            array_collections.update(self._train_diagnostics_routine.array_collections)
        if self._loader_routine is not None:
            array_collections.update(self._loader_routine.array_collections)
        if self._artifact_routine is not None:
            array_collections.update(self._artifact_routine.array_collections)
        return array_collections

    @property
    def plot_collections(self) -> Dict[str, PlotCollection]:
        plot_collections = {}
        if self._train_diagnostics_routine is not None:
            plot_collections.update(self._train_diagnostics_routine.plot_collections)
        if self._loader_routine is not None:
            plot_collections.update(self._loader_routine.plot_collections)
        if self._artifact_routine is not None:
            plot_collections.update(self._artifact_routine.plot_collections)
        return plot_collections

    def attach_train_diagnostics_hooks(self, model: ModelTContr, n_epochs_elapsed: int):
        if self._train_diagnostics_routine is not None:
            self._train_diagnostics_routine.attach(model=model, n_epochs_elapsed=n_epochs_elapsed)

    def execute(self, model: ModelTContr, n_epochs_elapsed: int):
        resources = RoutineResources(model=model, n_epochs_elapsed=n_epochs_elapsed)
        if self._train_diagnostics_routine is not None:
            self._train_diagnostics_routine.execute(resources=resources)
        if self._loader_routine is not None:
            self._loader_routine.execute(resources=resources)
        if self._artifact_routine is not None:
            self._artifact_routine.execute(resources=resources)

    def clear_cache(self):
        if self._train_diagnostics_routine is not None:
            self._train_diagnostics_routine.clear_cache()
        if self._loader_routine is not None:
            self._loader_routine.clear_cache()
        if self._artifact_routine is not None:
            self._artifact_routine.clear_cache()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Generic, Optional, TypeVar

import torch
from torch import optim
from tqdm import tqdm

from artifact_torch._base.data.data_loader import DataLoader
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._base.trainer.training_state import TrainingState

ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)


class TrainerBase(ABC, Generic[ModelTContr, ModelInputTContr, ModelOutputTContr]):
    def __init__(
        self,
        training_state: TrainingState[ModelTContr],
        train_loader: DataLoader[ModelInputTContr],
        device: torch.device,
    ):
        self._training_state = training_state
        self._train_loader = train_loader
        self._device = device
        self._n_epochs_elapsed = 0

    @property
    def training_state(self) -> TrainingState[ModelTContr]:
        return self._training_state

    @property
    def train_loader(self) -> DataLoader[ModelInputTContr]:
        return self._train_loader

    @property
    def optimizer(self) -> optim.Optimizer:
        return self._training_state.optimizer

    @property
    def scheduler(self) -> Optional[optim.lr_scheduler._LRScheduler]:
        return self._training_state.scheduler

    @property
    def device(self) -> torch.device:
        return self._device

    @device.setter
    def device(self, device: torch.device):
        self._device = device

    @property
    def n_epochs_elapsed(self) -> int:
        return self._n_epochs_elapsed

    @abstractmethod
    def _should_stop(self) -> bool: ...

    @abstractmethod
    def _execute_epoch_preprocessing(self): ...

    @abstractmethod
    def _execute_epoch_postprocessing(self):
        self._n_epochs_elapsed += 1

    @abstractmethod
    def _get_progress_bar_description(self) -> str: ...

    def train(self):
        self._training_state.model.device = self.device
        print(f"Training on device: {self.device}")
        while not self._should_stop():
            self._execute_epoch_preprocessing()
            self._train_one_epoch()
            self._execute_epoch_postprocessing()

    def load_checkpoint(self, path: Path):
        self._training_state.load(path=path, map_location=self.device)

    def _train_one_epoch(self):
        self._training_state.model.train()
        self._train_loader.device = self._training_state.model.device
        progress_bar_desc = self._get_progress_bar_description()
        for batch in tqdm(self._train_loader, desc=progress_bar_desc, leave=False):
            self._train_one_batch(batch=batch)
        self._training_state.model.eval()

    def _train_one_batch(self, batch: ModelInputTContr):
        model_output = self._training_state.model(batch)
        t_loss = self._get_loss(model_output=model_output)
        self._training_step(
            t_loss=t_loss,
            optimizer=self._training_state.optimizer,
            scheduler=self._training_state.scheduler,
        )
        return model_output

    @staticmethod
    def _training_step(
        t_loss: torch.Tensor,
        optimizer: optim.Optimizer,
        scheduler: Optional[optim.lr_scheduler._LRScheduler] = None,
    ):
        optimizer.zero_grad()
        t_loss.backward()
        optimizer.step()
        if scheduler is not None:
            scheduler.step()

    @staticmethod
    def _get_loss(model_output: ModelOutputTContr) -> torch.Tensor:
        loss = model_output.get("t_loss")
        assert loss is not None
        return loss

    def _get_checkpoint(self) -> Dict[str, Any]:
        return self._training_state.serialize()



from collections import defaultdict
from enum import Enum
from typing import Dict, Generic, List, Optional, TypeVar


class CacheMode(Enum):
    STANDARD = "STANDARD"
    ALIGNED = "ALIGNED"


CacheDataT = TypeVar("CacheDataT")


class Cache(Generic[CacheDataT]):
    def __init__(self, mode: CacheMode = CacheMode.STANDARD):
        self._mode = mode
        self._cache: Dict[str, List[Optional[CacheDataT]]] = defaultdict(list)

    @property
    def keys(self) -> List[str]:
        return list(self._cache.keys())

    @property
    def n_entries(self) -> int:
        if not self._cache:
            return 0
        first_key = next(iter(self._cache))
        return len(self._cache[first_key])

    @property
    def is_empty(self) -> bool:
        return self.n_entries == 0

    def __getitem__(self, key: str) -> Optional[List[Optional[CacheDataT]]]:
        return self._cache.get(key)

    def __repr__(self) -> str:
        return f"Cache({list(self._cache.keys())})"

    def get_full_history(self, key: str) -> Optional[List[Optional[CacheDataT]]]:
        return self._cache[key]

    def get_latest(self, key: str, default: Optional[CacheDataT] = None) -> Optional[CacheDataT]:
        history = self._cache.get(key)
        if history is None:
            latest = default
        else:
            latest = history[-1]
        return latest

    def clear(self) -> None:
        self._cache.clear()

    def append(self, items: Dict[str, CacheDataT]):
        if self._mode == CacheMode.STANDARD:
            self._append_standard(items=items)
        elif self._mode == CacheMode.ALIGNED:
            self._append_aligned(items=items)
        else:
            raise ValueError(f"Invalid cache mode: {self._mode}")

    def append_standard(self, items: Dict[str, CacheDataT]):
        for key, value in items.items():
            self._cache[key].append(value)

    def _append_standard(self, items: Dict[str, CacheDataT]):
        for key, value in items.items():
            self._cache[key].append(value)

    def _append_aligned(self, items: Dict[str, CacheDataT]):
        ls_new_keys = [key for key in items.keys() if key not in self._cache.keys()]
        for key in ls_new_keys:
            self._add_empty_history(cache=self._cache, key=key, n_entries=self.n_entries)
        self._add_items_aligned(cache=self._cache, items=items)

    @staticmethod
    def _add_items_aligned(
        cache: Dict[str, List[Optional[CacheDataT]]], items: Dict[str, CacheDataT]
    ):
        for key, col_list in cache.items():
            if key in items:
                col_list.append(items[key])
            else:
                col_list.append(None)

    @staticmethod
    def _add_empty_history(cache: Dict[str, List[Optional[CacheDataT]]], key: str, n_entries: int):
        cache[key] = [None] * n_entries


class StandardCache(Cache[CacheDataT]):
    def __init__(self):
        super().__init__(mode=CacheMode.STANDARD)


class AlignedCache(Cache[CacheDataT]):
    def __init__(self):
        super().__init__(mode=CacheMode.ALIGNED)



from artifact_torch._base.components.early_stopping.patience import (
    PatienceStopper,
    PatienceStopperUpdateData,
)
from artifact_torch._base.components.early_stopping.single_score import SingleScoreStopper
from artifact_torch._base.components.early_stopping.stopper import EarlyStopper, StopperUpdateData
from artifact_torch._impl.early_stopping.epoch_bound import EpochBoundStopper
from artifact_torch._impl.early_stopping.score_improvement import (
    ScoreMaximizationStopper,
    ScoreMinimizationStopper,
)

__all__ = [
    "PatienceStopper",
    "PatienceStopperUpdateData",
    "SingleScoreStopper",
    "EarlyStopper",
    "StopperUpdateData",
    "EpochBoundStopper",
    "ScoreMaximizationStopper",
    "ScoreMinimizationStopper",
]



from dataclasses import dataclass
from typing import Generic, TypeVar

from artifact_torch._base.components.callbacks.periodic import PeriodicTrackingCallbackResources
from artifact_torch._base.model.base import Model

ModelTCov = TypeVar("ModelTCov", bound=Model, covariant=True)


@dataclass(frozen=True)
class HookCallbackResources(PeriodicTrackingCallbackResources, Generic[ModelTCov]):
    model: ModelTCov



from abc import ABC, abstractmethod
from typing import Generic, Iterator, TypeVar

from torch.utils.data import Dataset as NativeDataset
from torch.utils.data import IterableDataset as NativeIterableDataset

from artifact_torch._base.model.io import ModelInput

ModelInputTCov = TypeVar("ModelInputTCov", bound=ModelInput, covariant=True)


class Dataset(NativeDataset, ABC, Generic[ModelInputTCov]):
    @abstractmethod
    def __len__(self) -> int: ...

    @abstractmethod
    def __getitem__(self, idx: int) -> ModelInputTCov: ...


class IterableDataset(NativeIterableDataset, ABC, Generic[ModelInputTCov]):
    @abstractmethod
    def __iter__(self) -> Iterator[ModelInputTCov]: ...



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, List, Mapping, Optional, Set, Type, TypeVar

from artifact_core.spi.resources import ArtifactResources, ResourceSpecProtocol
from artifact_experiment.spi.plans import ArtifactPlan, ArtifactPlanBuildContext
from artifact_experiment.tracking import DataSplit
from artifact_experiment.tracking.spi import TrackingQueue

from artifact_torch._base.components.routines.routine import PlanExecutionRoutine, RoutineResources
from artifact_torch._base.model.base import Model
from artifact_torch._utils.collections.key_selector import KeySelector
from artifact_torch._utils.scheduling.periodic_actions import PeriodicActionTrigger

ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)
ArtifactRoutineHyperparamsTCov = TypeVar(
    "ArtifactRoutineHyperparamsTCov", bound="ArtifactRoutineHyperparams", covariant=True
)
ArtifactRoutineDataTContr = TypeVar(
    "ArtifactRoutineDataTContr", bound="ArtifactRoutineData", contravariant=True
)
ArtifactResourcesTCov = TypeVar("ArtifactResourcesTCov", bound=ArtifactResources, covariant=True)
ResourceSpecProtocolTContr = TypeVar(
    "ResourceSpecProtocolTContr", bound=ResourceSpecProtocol, contravariant=True
)
ExportTContr = TypeVar("ExportTContr", contravariant=True)
ArtifactRoutineT = TypeVar("ArtifactRoutineT", bound="ArtifactRoutine[Any, Any, Any, Any, Any]")


@dataclass
class ArtifactRoutineHyperparams: ...


@dataclass
class ArtifactRoutineData: ...


class ArtifactRoutine(
    PlanExecutionRoutine[ModelTContr],
    Generic[
        ModelTContr,
        ArtifactRoutineDataTContr,
        ArtifactRoutineHyperparamsTCov,
        ArtifactResourcesTCov,
        ResourceSpecProtocolTContr,
    ],
):
    def __init__(
        self,
        data: Mapping[DataSplit, ArtifactRoutineDataTContr],
        periods: Mapping[DataSplit, int],
        artifact_plans: Mapping[
            DataSplit,
            ArtifactPlan[
                ArtifactResourcesTCov,
                ResourceSpecProtocolTContr,
                Any,
                Any,
                Any,
                Any,
                Any,
                Any,
                Any,
            ],
        ],
        hyperparams: ArtifactRoutineHyperparamsTCov,
    ):
        periods = KeySelector.restrict_to_keys(periods, keys_from=data)
        artifact_plans = KeySelector.restrict_to_keys(artifact_plans, keys_from=data)
        super().__init__(plans=list(artifact_plans.values()))
        self._data = data
        self._periods = periods
        self._artifact_plans = artifact_plans
        self._hyperparams = hyperparams

    @classmethod
    def build(
        cls: Type[ArtifactRoutineT],
        data: Mapping[DataSplit, ArtifactRoutineDataTContr],
        data_spec: ResourceSpecProtocolTContr,
        tracking_queue: Optional[TrackingQueue] = None,
    ) -> ArtifactRoutineT:
        plan_build_context = ArtifactPlanBuildContext(
            tracking_queue=tracking_queue, resource_spec=data_spec
        )
        periods = {
            data_split: period
            for data_split in DataSplit
            if (period := cls._get_period(data_split=data_split)) is not None
        }
        artifact_plans = {
            data_split: plan
            for data_split in DataSplit
            if (plan := cls._build_artifact_plan(data_split=data_split, context=plan_build_context))
            is not None
        }
        hyperparams = cls._get_hyperparams()
        routine = cls(
            data=data, periods=periods, artifact_plans=artifact_plans, hyperparams=hyperparams
        )
        return routine

    @property
    def _data_splits(self) -> List[DataSplit]:
        return list(self._data.keys())

    @classmethod
    @abstractmethod
    def _get_period(cls, data_split: DataSplit) -> Optional[int]: ...

    @classmethod
    @abstractmethod
    def _get_artifact_plan(
        cls, data_split: DataSplit
    ) -> Optional[
        Type[
            ArtifactPlan[
                ArtifactResourcesTCov,
                ResourceSpecProtocolTContr,
                Any,
                Any,
                Any,
                Any,
                Any,
                Any,
                Any,
            ]
        ]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_hyperparams(cls) -> ArtifactRoutineHyperparamsTCov: ...

    @abstractmethod
    def _generate_artifact_resources(
        self,
        model: ModelTContr,
    ) -> Mapping[DataSplit, ArtifactResourcesTCov]: ...

    def clear_cache(self):
        for validation_plan in self._artifact_plans.values():
            validation_plan.clear_cache()

    def execute(self, resources: RoutineResources[ModelTContr]):
        data_splits = self._get_active_splits(n_epochs_elapsed=resources.n_epochs_elapsed)
        artifact_resources_by_split = self._generate_artifact_resources(model=resources.model)
        for data_split in data_splits:
            artifact_plan = self._artifact_plans[data_split]
            artifact_resources = artifact_resources_by_split[data_split]
            artifact_plan.execute_artifacts(resources=artifact_resources, data_split=data_split)

    def _get_active_splits(self, n_epochs_elapsed: int) -> Set[DataSplit]:
        return {
            split
            for split in self._data_splits
            if PeriodicActionTrigger.should_trigger(
                step=n_epochs_elapsed, period=self._periods[split]
            )
        }

    @classmethod
    def _build_artifact_plan(
        cls, data_split: DataSplit, context: ArtifactPlanBuildContext[ResourceSpecProtocolTContr]
    ) -> Optional[
        ArtifactPlan[
            ArtifactResourcesTCov, ResourceSpecProtocolTContr, Any, Any, Any, Any, Any, Any, Any
        ]
    ]:
        plan_class = cls._get_artifact_plan(data_split=data_split)
        if plan_class is not None:
            return plan_class.build(context=context)



from abc import abstractmethod
from collections import deque
from dataclasses import dataclass
from typing import Generic, List, Optional, TypeVar

from artifact_torch._base.components.early_stopping.stopper import EarlyStopper, StopperUpdateData

PatienceStopperCriterionT = TypeVar("PatienceStopperCriterionT")


@dataclass
class PatienceStopperUpdateData(StopperUpdateData, Generic[PatienceStopperCriterionT]):
    criterion: PatienceStopperCriterionT


class PatienceStopper(
    EarlyStopper[PatienceStopperUpdateData[PatienceStopperCriterionT]],
    Generic[PatienceStopperCriterionT],
):
    def __init__(self, patience: int, max_n_epochs: Optional[int] = None):
        super().__init__(max_n_epochs=max_n_epochs)
        self._patience = patience
        self._criterion_history = deque[PatienceStopperCriterionT](maxlen=patience + 1)

    @property
    def patience(self) -> int:
        return self._patience

    @property
    def criterion_history(self) -> List[PatienceStopperCriterionT]:
        return list(self._criterion_history)

    @property
    def latest_criterion(self) -> Optional[PatienceStopperCriterionT]:
        if not self._queue_is_empty():
            return self._criterion_history[-1]

    @property
    def earliest_criterion(self) -> Optional[PatienceStopperCriterionT]:
        if not self._queue_is_empty():
            return self._criterion_history[0]

    @abstractmethod
    def stopping_condition_met(self) -> bool:
        pass

    def _update(self, update_data: PatienceStopperUpdateData[PatienceStopperCriterionT]):
        super()._update(update_data=update_data)
        self._criterion_history.append(update_data.criterion)
        self._current_criterion = update_data.criterion

    def _queue_is_empty(self) -> bool:
        return len(self._criterion_history) == 0

    def _queue_is_full(self) -> bool:
        return len(self._criterion_history) == self._patience + 1



from typing import Generic, Iterator, TypeVar, Union

import torch
from torch.utils.data import DataLoader as NativeDataLoader

from artifact_torch._base.data.dataset import Dataset, IterableDataset
from artifact_torch._base.data.device_manager import DeviceManager
from artifact_torch._base.model.io import ModelInput

ModelInputTCov = TypeVar("ModelInputTCov", bound=ModelInput, covariant=True)


class DataLoader(NativeDataLoader, Generic[ModelInputTCov]):
    def __init__(
        self,
        dataset: Union[Dataset[ModelInputTCov], IterableDataset[ModelInputTCov]],
        *args,
        **kwargs,
    ):
        super().__init__(dataset, *args, **kwargs)
        self._device = torch.device("cpu")

    def __iter__(self) -> Iterator[ModelInputTCov]:  # type: ignore
        for batch in super().__iter__():
            batch = DeviceManager.move_to_device(data=batch, device=self._device)
            yield batch

    @property
    def device(self) -> torch.device:
        return self._device

    @device.setter
    def device(self, device: torch.device):
        self._device = device



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Mapping, Sequence, TypeVar

from artifact_core.typing import (
    Array,
    ArrayCollection,
    Plot,
    PlotCollection,
    Score,
    ScoreCollection,
)
from artifact_experiment.spi.plans import CallbackExecutionPlan

from artifact_torch._base.model.base import Model

ModelTCov = TypeVar("ModelTCov", bound=Model[Any, Any], covariant=True)


@dataclass(frozen=True)
class RoutineResources(Generic[ModelTCov]):
    model: ModelTCov
    n_epochs_elapsed: int


ModelTContr = TypeVar("ModelTContr", bound=Model[Any, Any], contravariant=True)


class PlanExecutionRoutine(ABC, Generic[ModelTContr]):
    def __init__(self, plans: Sequence[CallbackExecutionPlan]):
        self._plans = plans

    @property
    def scores(self) -> Mapping[str, Score]:
        scores = {}
        for plan in self._plans:
            scores.update(plan.scores)
        return scores

    @property
    def arrays(self) -> Mapping[str, Array]:
        arrays = {}
        for plan in self._plans:
            arrays.update(plan.arrays)
        return arrays

    @property
    def plots(self) -> Mapping[str, Plot]:
        plots = {}
        for plan in self._plans:
            plots.update(plan.plots)
        return plots

    @property
    def score_collections(self) -> Mapping[str, ScoreCollection]:
        score_collections = {}
        for plan in self._plans:
            score_collections.update(plan.score_collections)
        return score_collections

    @property
    def array_collections(self) -> Mapping[str, ArrayCollection]:
        array_collections = {}
        for plan in self._plans:
            array_collections.update(plan.array_collections)
        return array_collections

    @property
    def plot_collections(self) -> Mapping[str, PlotCollection]:
        plot_collections = {}
        for plan in self._plans:
            plot_collections.update(plan.plot_collections)
        return plot_collections

    @abstractmethod
    def execute(self, resources: RoutineResources[ModelTContr]): ...

    def clear_cache(self):
        for plan in self._plans:
            plan.clear_cache()



from pathlib import Path


class FilenameModifier:
    @staticmethod
    def append(filepath: str | Path, text: str) -> Path:
        p = Path(filepath)
        return p.with_name(f"{p.stem}{text}{p.suffix}")

    @staticmethod
    def replace(filepath: str | Path, new_basename: str) -> Path:
        p = Path(filepath)
        return p.with_name(f"{new_basename}{p.suffix}")



from abc import abstractmethod
from typing import Generic, Optional, TypeVar

from artifact_experiment.spi.callbacks import CacheCallback, Callback, TrackingCallback
from artifact_experiment.tracking.spi import TrackingQueueWriter
from artifact_experiment.typing import TrackingData

from artifact_torch._base.components.resources.periodic import (
    PeriodicCacheCallbackResources,
    PeriodicCallbackResources,
    PeriodicTrackingCallbackResources,
)
from artifact_torch._utils.scheduling.periodic_actions import PeriodicActionTrigger

PeriodicCallbackResourcesTContr = TypeVar(
    "PeriodicCallbackResourcesTContr", bound=PeriodicCallbackResources, contravariant=True
)


class PeriodicCallback(
    Callback[PeriodicCallbackResourcesTContr],
    Generic[PeriodicCallbackResourcesTContr],
):
    def __init__(self, period: int):
        self._period = period

    @property
    def execution_interval(self) -> int:
        return self._period

    @abstractmethod
    def _execute(self, resources: PeriodicCallbackResourcesTContr): ...

    def execute(self, resources: PeriodicCallbackResourcesTContr):
        should_trigger = self._should_trigger(step=resources.step)
        if should_trigger:
            self._execute(resources=resources)

    def _should_trigger(self, step: int):
        should_trigger = PeriodicActionTrigger.should_trigger(step=step, period=self._period)
        return should_trigger


PeriodicCacheCallbackResourcesTContr = TypeVar(
    "PeriodicCacheCallbackResourcesTContr", bound=PeriodicCacheCallbackResources, contravariant=True
)
CacheDataTCov = TypeVar("CacheDataTCov", bound=TrackingData, covariant=True)


class PeriodicCacheCallback(
    CacheCallback[PeriodicCacheCallbackResourcesTContr, CacheDataTCov],
    Generic[PeriodicCacheCallbackResourcesTContr, CacheDataTCov],
):
    def __init__(self, base_key: str, period: int):
        super().__init__(base_key=base_key)
        self._period = period

    @abstractmethod
    def _compute(self, resources: PeriodicCacheCallbackResourcesTContr) -> CacheDataTCov: ...

    @property
    def execution_interval(self) -> int:
        return self._period

    def execute(self, resources: PeriodicCacheCallbackResourcesTContr):
        self._clear()
        should_trigger = self._should_trigger(step=resources.step)
        if should_trigger:
            super().execute(resources=resources)

    def _should_trigger(self, step: int):
        should_trigger = PeriodicActionTrigger.should_trigger(step=step, period=self._period)
        return should_trigger


PeriodicTrackingCallbackResourcesTContr = TypeVar(
    "PeriodicTrackingCallbackResourcesTContr",
    bound=PeriodicTrackingCallbackResources,
    contravariant=True,
)


class PeriodicTrackingCallback(
    TrackingCallback[PeriodicTrackingCallbackResourcesTContr, CacheDataTCov],
    Generic[PeriodicTrackingCallbackResourcesTContr, CacheDataTCov],
):
    def __init__(
        self,
        base_key: str,
        period: int,
        writer: Optional[TrackingQueueWriter[CacheDataTCov]] = None,
    ):
        super().__init__(base_key=base_key, writer=writer)
        self._period = period

    @abstractmethod
    def _compute(self, resources: PeriodicTrackingCallbackResourcesTContr) -> CacheDataTCov: ...

    @property
    def execution_interval(self) -> int:
        return self._period

    def execute(self, resources: PeriodicTrackingCallbackResourcesTContr):
        self._clear()
        should_trigger = self._should_trigger(step=resources.step)
        if should_trigger:
            super().execute(resources=resources)

    def _should_trigger(self, step: int):
        should_trigger = PeriodicActionTrigger.should_trigger(step=step, period=self._period)
        return should_trigger



from artifact_torch._base.components.callbacks.model_io import (
    ModelIOArrayCallback,
    ModelIOArrayCollectionCallback,
    ModelIOCallback,
    ModelIOPlotCallback,
    ModelIOPlotCollectionCallback,
    ModelIOScoreCallback,
    ModelIOScoreCollectionCallback,
)
from artifact_torch._impl.callbacks.model_io.loss import LossCallback

__all__ = [
    "ModelIOArrayCallback",
    "ModelIOArrayCollectionCallback",
    "ModelIOCallback",
    "ModelIOPlotCallback",
    "ModelIOPlotCollectionCallback",
    "ModelIOScoreCallback",
    "ModelIOScoreCollectionCallback",
    "LossCallback",
]



from artifact_torch._base.components.plans.backward_hook import (
    BackwardHookPlan,
    BackwardHookPlanBuildContext,
)
from artifact_torch._base.components.plans.forward_hook import (
    ForwardHookPlan,
    ForwardHookPlanBuildContext,
)
from artifact_torch._base.components.plans.model_io import ModelIOPlan, ModelIOPlanBuildContext

__all__ = [
    "BackwardHookPlan",
    "BackwardHookPlanBuildContext",
    "ForwardHookPlan",
    "ForwardHookPlanBuildContext",
    "ModelIOPlan",
    "ModelIOPlanBuildContext",
]



from typing import Any, Generic, TypeVar

from artifact_core.typing import (
    Array,
    ArrayCollection,
    ArtifactResult,
    Plot,
    PlotCollection,
    Score,
    ScoreCollection,
)
from artifact_experiment.spi.handlers import TrackingCallbackHandler

from artifact_torch._base.components.callbacks.hook import HookCallback, HookCallbackResources
from artifact_torch._base.model.base import Model

ModelTContr = TypeVar("ModelTContr", bound=Model, contravariant=True)
CacheDataTCov = TypeVar("CacheDataTCov", bound=ArtifactResult, covariant=True)


class HookCallbackHandler(
    TrackingCallbackHandler[
        HookCallback[ModelTContr, CacheDataTCov, Any],
        HookCallbackResources[ModelTContr],
        CacheDataTCov,
    ],
    Generic[ModelTContr, CacheDataTCov],
):
    def attach(self, resources: HookCallbackResources[ModelTContr]) -> bool:
        any_attached = False
        for callback in self._ls_callbacks:
            any_attached |= callback.attach(resources=resources)
        return any_attached


HookScoreHandler = HookCallbackHandler[ModelTContr, Score]


HookArrayHandler = HookCallbackHandler[ModelTContr, Array]


HookPlotHandler = HookCallbackHandler[ModelTContr, Plot]


HookScoreCollectionHandler = HookCallbackHandler[ModelTContr, ScoreCollection]


HookArrayCollectionHandler = HookCallbackHandler[ModelTContr, ArrayCollection]


HookPlotCollectionHandler = HookCallbackHandler[ModelTContr, PlotCollection]



from artifact_torch._base.components.early_stopping.stopper import EarlyStopper, StopperUpdateData


class EpochBoundStopper(EarlyStopper[StopperUpdateData]):
    def __init__(self, max_n_epochs: int):
        super().__init__(max_n_epochs=max_n_epochs)

    def stopping_condition_met(self) -> bool:
        max_epochs_exceeded = self._max_epochs_exceeded()
        return max_epochs_exceeded



from artifact_torch._base.components.routines.loader import DataLoaderRoutine
from artifact_torch._base.components.routines.train_diagnostics import TrainDiagnosticsRoutine

__all__ = ["DataLoaderRoutine", "TrainDiagnosticsRoutine"]



from collections.abc import Sequence
from typing import Any, Generic, Type, TypeVar

from artifact_experiment.spi.handler_suites import TrackingCallbackHandlerSuite

from artifact_torch._base.components.callbacks.hook import (
    HookArrayCallback,
    HookArrayCollectionCallback,
    HookCallback,
    HookPlotCallback,
    HookPlotCollectionCallback,
    HookScoreCallback,
    HookScoreCollectionCallback,
)
from artifact_torch._base.components.handlers.hook import HookCallbackHandler
from artifact_torch._base.components.resources.hook import HookCallbackResources
from artifact_torch._base.model.base import Model

ModelTContr = TypeVar("ModelTContr", bound=Model, contravariant=True)
HookCallbackHandlerSuiteT = TypeVar(
    "HookCallbackHandlerSuiteT", bound="HookCallbackHandlerSuite[Any]"
)


class HookCallbackHandlerSuite(
    TrackingCallbackHandlerSuite[
        HookCallbackHandler[ModelTContr, Any],
        HookCallback[ModelTContr, Any, Any],
        HookCallbackResources[ModelTContr],
    ],
    Generic[ModelTContr],
):
    @classmethod
    def build(
        cls: Type[HookCallbackHandlerSuiteT],
        score_callbacks: Sequence[HookScoreCallback[ModelTContr]],
        array_callbacks: Sequence[HookArrayCallback[ModelTContr]],
        plot_callbacks: Sequence[HookPlotCallback[ModelTContr]],
        score_collection_callbacks: Sequence[HookScoreCollectionCallback[ModelTContr]],
        array_collection_callbacks: Sequence[HookArrayCollectionCallback[ModelTContr]],
        plot_collection_callbacks: Sequence[HookPlotCollectionCallback[ModelTContr]],
    ) -> HookCallbackHandlerSuiteT:
        handler_suite = cls(
            score_handler=HookCallbackHandler(callbacks=score_callbacks),
            array_handler=HookCallbackHandler(callbacks=array_callbacks),
            plot_handler=HookCallbackHandler(callbacks=plot_callbacks),
            score_collection_handler=HookCallbackHandler(callbacks=score_collection_callbacks),
            array_collection_handler=HookCallbackHandler(callbacks=array_collection_callbacks),
            plot_collection_handler=HookCallbackHandler(callbacks=plot_collection_callbacks),
        )
        return handler_suite

    def attach(self, resources: HookCallbackResources[ModelTContr]) -> bool:
        any_attached = False
        for handler in self._ls_handlers:
            any_attached |= handler.attach(resources=resources)
        return any_attached



from abc import abstractmethod
from typing import Any, Generic, Optional, Type, TypeVar

from artifact_core.binary_classification.spi import BinaryClassSpecProtocol

from artifact_torch._base.components.routines.loader import DataLoaderRoutine
from artifact_torch._base.components.routines.train_diagnostics import TrainDiagnosticsRoutine
from artifact_torch._base.experiment.experiment import Experiment
from artifact_torch._base.model.io import ModelInput, ModelOutput
from artifact_torch._base.trainer.trainer import Trainer
from artifact_torch._domains.classification.model import ClassificationParams
from artifact_torch.binary_classification._model import BinaryClassifier
from artifact_torch.binary_classification._routine import (
    BinaryClassificationRoutine,
    BinaryClassificationRoutineData,
)

BinaryClassifierTContr = TypeVar(
    "BinaryClassifierTContr", bound=BinaryClassifier[Any, Any, Any, Any]
)
ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput)
ClassificationParamsTContr = TypeVar("ClassificationParamsTContr", bound=ClassificationParams)
ClassificationDataTContr = TypeVar("ClassificationDataTContr")
BinaryClassificationExperimentT = TypeVar(
    "BinaryClassificationExperimentT", bound="BinaryClassificationExperiment"
)


class BinaryClassificationExperiment(
    Experiment[
        BinaryClassifierTContr,
        ModelInputTContr,
        ModelOutputTContr,
        BinaryClassificationRoutineData[ClassificationDataTContr],
        BinaryClassSpecProtocol,
    ],
    Generic[
        BinaryClassifierTContr,
        ModelInputTContr,
        ModelOutputTContr,
        ClassificationParamsTContr,
        ClassificationDataTContr,
    ],
):
    @classmethod
    @abstractmethod
    def _get_trainer(
        cls,
    ) -> Type[
        Trainer[
            BinaryClassifierTContr,
            ModelInputTContr,
            ModelOutputTContr,
            Any,
            Any,
        ]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_train_diagnostics_routine(
        cls,
    ) -> Optional[
        Type[TrainDiagnosticsRoutine[BinaryClassifierTContr, ModelInputTContr, ModelOutputTContr]]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_loader_routine(
        cls,
    ) -> Optional[
        Type[DataLoaderRoutine[BinaryClassifierTContr, ModelInputTContr, ModelOutputTContr]]
    ]: ...

    @classmethod
    @abstractmethod
    def _get_artifact_routine(
        cls,
    ) -> Optional[
        Type[BinaryClassificationRoutine[ClassificationParamsTContr, ClassificationDataTContr]]
    ]: ...



from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from artifact_torch._base.components.cache.cache import AlignedCache


class ScoreCache(AlignedCache[float]):
    def __init__(self):
        super().__init__()
        self._df_scores: pd.DataFrame = self._build_df(self._cache)

    @property
    def scores(self) -> pd.DataFrame:
        return self._df_scores

    def clear(self) -> None:
        super().clear()
        self._df_scores = self._build_df(cache=self._cache)

    def append(self, items: Dict[str, float]) -> None:
        super().append(items)
        self._df_scores = self._build_df(self._cache)

    def get_latest_non_null(self, key: str) -> Optional[float]:
        if key in self._df_scores.columns:
            series = self._df_scores[key].dropna()
            if not series.empty:
                return float(series.iloc[-1])

    @staticmethod
    def _build_df(cache: Dict[str, List[Optional[float]]]) -> pd.DataFrame:
        sanitized: Dict[str, List[float]] = {
            k: [np.nan if v_item is None else v_item for v_item in v_list]
            for k, v_list in cache.items()
        }
        df = pd.DataFrame(sanitized).astype(pd.SparseDtype("float"))
        return df



from typing import Any, Dict, List, Tuple, overload

import torch

from artifact_torch._base.model.io import ModelIO


class DeviceManager:
    @classmethod
    @overload
    def move_to_device(cls, data: ModelIO, device: torch.device) -> ModelIO: ...

    @classmethod
    @overload
    def move_to_device(cls, data: Dict[str, Any], device: torch.device) -> Dict[str, Any]: ...

    @classmethod
    @overload
    def move_to_device(cls, data: List[Any], device: torch.device) -> List[Any]: ...

    @classmethod
    @overload
    def move_to_device(cls, data: Tuple[Any, ...], device: torch.device) -> Tuple[Any, ...]: ...

    @classmethod
    @overload
    def move_to_device(cls, data: torch.Tensor, device: torch.device) -> torch.Tensor: ...

    @classmethod
    @overload
    def move_to_device(cls, data: Any, device: torch.device) -> Any: ...

    @classmethod
    def move_to_device(cls, data: Any, device: torch.device) -> Any:
        if isinstance(data, dict):
            return {k: cls.move_to_device(data=v, device=device) for k, v in data.items()}
        elif isinstance(data, list):
            return [cls.move_to_device(data=item, device=device) for item in data]
        elif isinstance(data, tuple):
            return tuple(cls.move_to_device(data=item, device=device) for item in data)
        elif isinstance(data, torch.Tensor):
            return data.to(device)
        else:
            return data



from collections.abc import Sequence
from typing import Any, Generic, Type, TypeVar

from artifact_experiment.spi.handler_suites import TrackingCallbackHandlerSuite

from artifact_torch._base.components.callbacks.model_io import (
    ModelIOArrayCallback,
    ModelIOArrayCollectionCallback,
    ModelIOCallback,
    ModelIOPlotCallback,
    ModelIOPlotCollectionCallback,
    ModelIOScoreCallback,
    ModelIOScoreCollectionCallback,
)
from artifact_torch._base.components.handlers.model_io import ModelIOCallbackHandler
from artifact_torch._base.components.resources.model_io import ModelIOCallbackResources
from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
ModelIOCallbackHandlerSuiteT = TypeVar(
    "ModelIOCallbackHandlerSuiteT", bound="ModelIOCallbackHandlerSuite[Any, Any]"
)


class ModelIOCallbackHandlerSuite(
    TrackingCallbackHandlerSuite[
        ModelIOCallbackHandler[ModelOutputTContr, Any],
        ModelIOCallback[ModelInputTContr, ModelOutputTContr, Any, Any],
        ModelIOCallbackResources[ModelOutputTContr],
    ],
    Generic[ModelInputTContr, ModelOutputTContr],
):
    @classmethod
    def build(
        cls: Type[ModelIOCallbackHandlerSuiteT],
        score_callbacks: Sequence[ModelIOScoreCallback[ModelInputTContr, ModelOutputTContr]],
        array_callbacks: Sequence[ModelIOArrayCallback[ModelInputTContr, ModelOutputTContr]],
        plot_callbacks: Sequence[ModelIOPlotCallback[ModelInputTContr, ModelOutputTContr]],
        score_collection_callbacks: Sequence[
            ModelIOScoreCollectionCallback[ModelInputTContr, ModelOutputTContr]
        ],
        array_collection_callbacks: Sequence[
            ModelIOArrayCollectionCallback[ModelInputTContr, ModelOutputTContr]
        ],
        plot_collection_callbacks: Sequence[
            ModelIOPlotCollectionCallback[ModelInputTContr, ModelOutputTContr]
        ],
    ) -> ModelIOCallbackHandlerSuiteT:
        handler_suite = cls(
            score_handler=ModelIOCallbackHandler(callbacks=score_callbacks),
            array_handler=ModelIOCallbackHandler(callbacks=array_callbacks),
            plot_handler=ModelIOCallbackHandler(callbacks=plot_callbacks),
            score_collection_handler=ModelIOCallbackHandler(callbacks=score_collection_callbacks),
            array_collection_handler=ModelIOCallbackHandler(callbacks=array_collection_callbacks),
            plot_collection_handler=ModelIOCallbackHandler(callbacks=plot_collection_callbacks),
        )
        return handler_suite

    def attach(self, resources: ModelIOCallbackResources[ModelOutputTContr]) -> bool:
        any_attached = False
        for handler in self._ls_handlers:
            any_attached |= handler.attach(resources=resources)
        return any_attached



from dataclasses import dataclass
from typing import Any, Mapping

from artifact_experiment.spi.resources import ExportCallbackResources


@dataclass(frozen=True)
class CheckpointCallbackResources(ExportCallbackResources[Mapping[str, Any]]):
    epoch: int



from abc import abstractmethod
from collections.abc import Sequence
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar

import torch
import torch.nn as nn
from artifact_core.typing import (
    Array,
    ArrayCollection,
    ArtifactResult,
    Plot,
    PlotCollection,
    Score,
    ScoreCollection,
)

from artifact_torch._base.components.callbacks.forward_hook import ForwardHookCallback
from artifact_torch._base.model.base import Model
from artifact_torch._base.model.io import ModelInput, ModelOutput

ModelInputTContr = TypeVar("ModelInputTContr", bound=ModelInput, contravariant=True)
ModelOutputTContr = TypeVar("ModelOutputTContr", bound=ModelOutput, contravariant=True)
CacheDataT = TypeVar("CacheDataT", bound=ArtifactResult, covariant=True)
BatchResultT = TypeVar("BatchResultT")


class ModelIOCallback(
    ForwardHookCallback[Model[Any, ModelOutputTContr], CacheDataT, BatchResultT],
    Generic[ModelInputTContr, ModelOutputTContr, CacheDataT, BatchResultT],
):
    @classmethod
    @abstractmethod
    def _get_base_key(cls) -> str: ...

    @classmethod
    @abstractmethod
    def _compute_on_batch(
        cls, model_input: ModelInputTContr, model_output: ModelOutputTContr
    ) -> BatchResultT: ...

    @classmethod
    @abstractmethod
    def _aggregate_batch_results(cls, ls_batch_results: List[BatchResultT]) -> CacheDataT: ...

    @classmethod
    def _get_layers(cls, model: Model[ModelInputTContr, ModelOutputTContr]) -> Sequence[nn.Module]:
        return [model]

    @classmethod
    def _hook(
        cls, module: nn.Module, inputs: Tuple[Any, ...], output: Any
    ) -> Optional[BatchResultT]:
        _ = module
        return cls._compute_on_batch(model_input=inputs[0], model_output=output)

    @classmethod
    def _aggregate(cls, hook_results: Dict[str, List[BatchResultT]]) -> CacheDataT:
        ls_batch_results = next(iter(hook_results.values()), [])
        return cls._aggregate_batch_results(ls_batch_results=ls_batch_results)


ModelIOScoreCallback = ModelIOCallback[
    ModelInputTContr, ModelOutputTContr, Score, Dict[str, torch.Tensor]
]


ModelIOArrayCallback = ModelIOCallback[
    ModelInputTContr, ModelOutputTContr, Array, Dict[str, torch.Tensor]
]


ModelIOPlotCallback = ModelIOCallback[
    ModelInputTContr, ModelOutputTContr, Plot, Dict[str, torch.Tensor]
]


ModelIOScoreCollectionCallback = ModelIOCallback[
    ModelInputTContr, ModelOutputTContr, ScoreCollection, Dict[str, torch.Tensor]
]


ModelIOArrayCollectionCallback = ModelIOCallback[
    ModelInputTContr, ModelOutputTContr, ArrayCollection, Dict[str, torch.Tensor]
]


ModelIOPlotCollectionCallback = ModelIOCallback[
    ModelInputTContr, ModelOutputTContr, PlotCollection, Dict[str, torch.Tensor]
]



from abc import abstractmethod
from typing import Dict, Generic, List, Optional, Sequence, TypeVar

import torch
import torch.nn as nn
from artifact_core.typing import (
    Array,
    ArrayCollection,
    ArtifactResult,
    Plot,
    PlotCollection,
    Score,
    ScoreCollection,
)
from artifact_experiment.tracking.spi import TrackingQueueWriter
from torch.utils.hooks import RemovableHandle

from artifact_torch._base.components.callbacks.periodic import PeriodicTrackingCallback
from artifact_torch._base.components.resources.hook import HookCallbackResources
from artifact_torch._base.model.base import Model

ModelTContr = TypeVar("ModelTContr", bound=Model, contravariant=True)
CacheDataTCov = TypeVar("CacheDataTCov", bound=ArtifactResult, covariant=True)
HookResultT = TypeVar("HookResultT")


class HookCallback(
    PeriodicTrackingCallback[HookCallbackResources[ModelTContr], CacheDataTCov],
    Generic[ModelTContr, CacheDataTCov, HookResultT],
):
    def __init__(self, period: int, writer: Optional[TrackingQueueWriter[CacheDataTCov]] = None):
        base_key = self._get_base_key()
        super().__init__(base_key=base_key, period=period, writer=writer)
        self._hook_results: Dict[str, List[HookResultT]] = {}
        self._handles: List[RemovableHandle] = []

    @classmethod
    @abstractmethod
    def _get_base_key(cls) -> str: ...

    @classmethod
    @abstractmethod
    def _get_layers(cls, model: ModelTContr) -> Sequence[nn.Module]: ...

    @classmethod
    @abstractmethod
    def _attach(
        cls, model: ModelTContr, sink: Dict[str, List[HookResultT]], handles: List[RemovableHandle]
    ): ...

    @classmethod
    @abstractmethod
    def _aggregate(cls, hook_results: Dict[str, List[HookResultT]]) -> CacheDataTCov: ...

    def attach(self, resources: HookCallbackResources[ModelTContr]) -> bool:
        if self._should_trigger(step=resources.step):
            self._attach(model=resources.model, sink=self._hook_results, handles=self._handles)
            return True
        else:
            return False

    def _compute(self, resources: HookCallbackResources[ModelTContr]) -> CacheDataTCov:
        _ = resources
        result = self._aggregate(hook_results=self._hook_results)
        self._hook_results.clear()
        self._detach()
        return result

    def _detach(self):
        for h in self._handles:
            h.remove()
        self._handles.clear()


HookScoreCallback = HookCallback[ModelTContr, Score, Dict[str, torch.Tensor]]

HookArrayCallback = HookCallback[ModelTContr, Array, Dict[str, torch.Tensor]]


HookPlotCallback = HookCallback[ModelTContr, Plot, Dict[str, torch.Tensor]]


HookScoreCollectionCallback = HookCallback[ModelTContr, ScoreCollection, Dict[str, torch.Tensor]]


HookArrayCollectionCallback = HookCallback[ModelTContr, ArrayCollection, Dict[str, torch.Tensor]]


HookPlotCollectionCallback = HookCallback[ModelTContr, PlotCollection, Dict[str, torch.Tensor]]
